/* 
 *  Matchbox Window Manager - A lightweight window manager not for the
 *                            desktop.
 *
 *  Authored By Matthew Allum <mallum@o-hand.com>
 *
 *  Copyright (c) 2002, 2004 OpenedHand Ltd - http://o-hand.com
 *
 *  SPDX-License-Identifier: GPL-2.0-or-later
 *
 */

#ifndef _WM_H_
#define _WM_H_

#include "structs.h"
#include "stack.h"
#include "base_client.h"
#include "main_client.h"
#include "toolbar_client.h"
#include "dockbar_client.h"
